package me.hammerle.snuviengine.api;

import java.io.File;
import java.io.IOException;
import java.nio.FloatBuffer;
import java.nio.file.Files;
import java.util.LinkedList;
import java.util.List;
import org.lwjgl.BufferUtils;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.*;

public final class Shader
{
    private static int program = -1;
    private static String[][] lightStrings;
    
    private static final List<Runnable> TASKS = new LinkedList<>();
    
    protected static boolean initDone = false;
    
    protected static void init()
    {
        program = createShaderProgram("shaders/vertex.vs", "shaders/fragment.fs");
        glUseProgram(program);
        
        sendMatrix();
        setCamera(0.0f, 0.0f);
        setDepth(0.0f);
        setAmbientLight(1.0f, 1.0f, 1.0f);
        
        lightStrings = new String[32][3];
        for(int index = 0; index < lightStrings.length; index++)
        {
            lightStrings[index][0] = "lights[" + index + "].color";
            lightStrings[index][1] = "lights[" + index + "].pos";
            lightStrings[index][2] = "lights[" + index + "].strength";
            
            setLightColor(index, 0.0f, 0.0f, 0.0f);
            setLightLocation(index, 0.0f, 0.0f);
            setLightStrength(index, 0.0f);
        }
        
        setTextureUsing(false);
        setLightUsing(false);
        setCameraUsing(false);
        
        initDone = true;
    }
    
    protected static void addTask(Runnable r)
    {
        TASKS.add(r);
    }
    
    protected static void doTasks()
    {
        if(!TASKS.isEmpty())
        {
            TASKS.forEach(r -> r.run());
            TASKS.clear();
        }
    }
    
    private static void sendMatrix()
    {
        FloatBuffer buffer = BufferUtils.createFloatBuffer(16);
        
        buffer.put(2.0f / Engine.WIDTH);
        buffer.put(0.0f);
        buffer.put(0.0f);
        buffer.put(0.0f);
        
        buffer.put(0.0f);
        buffer.put(2.0f / Engine.HEIGHT);
        buffer.put(0.0f);
        buffer.put(0.0f);
        
        buffer.put(0.0f);
        buffer.put(0.0f);
        buffer.put(1.0f);
        buffer.put(0.0f);
        
        buffer.put(-1.0f);
        buffer.put(-1.0f);
        buffer.put(0.0f);
        buffer.put(1.0f);
        
        buffer.flip();
        
        glUniformMatrix4fv(glGetUniformLocation(program, "matrix"), false, buffer);
    }
    
    public static void setCamera(float x, float y)
    {
        glUniform2f(glGetUniformLocation(program, "camera"), x, y);
    }
    
    public static void setDepth(float depth)
    {
        glUniform1f(glGetUniformLocation(program, "depth"), depth);
    }
    
    public static void setAmbientLight(float r, float g, float b)
    {
        glUniform3f(glGetUniformLocation(program, "ambientLight"), r, g, b);
    }
    
    private static void checkLightIndex(int index)
    {
        if(index < 0 || index > lightStrings.length)
        {
            throw new ShaderException("'" + index + "' is not a valid light index");
        }
    }
    
    public static void setLightColor(int index, float r, float g, float b)
    {
        checkLightIndex(index);
        glUniform3f(glGetUniformLocation(program, lightStrings[index][0]), r, g, b);
    }
    
    public static void setLightLocation(int index, float x, float y)
    {
        checkLightIndex(index);
        glUniform2f(glGetUniformLocation(program, lightStrings[index][1]), x, y);
    }
    
    public static void setLightStrength(int index, float strength)
    {
        checkLightIndex(index);
        glUniform1f(glGetUniformLocation(program, lightStrings[index][2]), strength);
    }
    
    public static void setTextureUsing(boolean use)
    {
        glUniform1i(glGetUniformLocation(program, "useTexture"), use ? 1 : 0);
    }
    
    public static void setLightUsing(boolean use)
    {
        glUniform1i(glGetUniformLocation(program, "useLight"), use ? 1 : 0);
    }
    
    public static void setCameraUsing(boolean use)
    {
        glUniform1i(glGetUniformLocation(program, "useCameraOffset"), use ? 1 : 0);
    }
    
    // -------------------------------------------------------------------------
    // general stuff
    // -------------------------------------------------------------------------
    
    private static String[] readFile(String path)
    {
        try
        {
            File f = new File(path);
            List<String> list = Files.readAllLines(f.toPath());
            list.replaceAll(s -> s + "\n");
            return list.toArray(new String[list.size()]);
        }
        catch(IOException ex)
        {
            throw new ShaderException("failed reading shader file '" + path + "'");
        }
    }
    
    private static String getError()
    {
        StringBuilder sb = new StringBuilder();
        
        int glErr = glGetError();
        while(glErr != GL_NO_ERROR)
        {
            sb.append(glErr);
            sb.append(" ");
            glErr = glGetError();
        }
        
        return sb.toString();
    }
    
    private static int createShaderProgram(String vertex, String fragment)
    {
        // ---------------------------------------------------------------------
        // vertex shader
        // ---------------------------------------------------------------------
        
        String vShaderSource[] = readFile(vertex);
        
        int vShader = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(vShader, vShaderSource);
        glCompileShader(vShader);

        String error = getError();
        if(!error.isEmpty())
        {
            throw new ShaderException("failed compiling vertex shader '" + vertex + "' " + error);
        }
        
        int compiled = glGetShaderi(vShader, GL_COMPILE_STATUS);
        if(compiled != 1)
        {
            throw new ShaderException("failed compiling vertex shader '" + vertex + "' " + compiled + " " + glGetShaderInfoLog(vShader));
        }
        
        // ---------------------------------------------------------------------
        // fragment shader
        // ---------------------------------------------------------------------

        String fShaderSource[] = readFile(fragment);
        
        int fShader = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(fShader, fShaderSource);
        glCompileShader(fShader);

        error = getError();
        if(!error.isEmpty())
        {
            throw new ShaderException("failed compiling fragment shader '" + fragment + "' " + error);
        }
        
        compiled = glGetShaderi(fShader, GL_COMPILE_STATUS);
        if(compiled != 1)
        {
            throw new ShaderException("failed compiling fragment shader '" + fragment + "' " + compiled + " " + glGetShaderInfoLog(fShader));
        }
        
        // ---------------------------------------------------------------------
        // linking
        // ---------------------------------------------------------------------

        int vfprogram = glCreateProgram();
        glAttachShader(vfprogram, vShader);
        glAttachShader(vfprogram, fShader);
        glLinkProgram(vfprogram);

        error = getError();
        if(!error.isEmpty())
        {
            throw new ShaderException("failed linking shaders '" + vertex + "' and '" + fragment + "' " + error);
        }
        
        compiled = glGetProgrami(vfprogram, GL_LINK_STATUS);
        if(compiled != 1)
        {
            throw new ShaderException("failed linking shaders '" + vertex + "' and '" + fragment + "' " + compiled);
        }

        glDeleteShader(vShader);
        glDeleteShader(fShader);
        
        return vfprogram;
    }  
}
